import json
from copy import deepcopy

from django.test import RequestFactory, TestCase
from mock import patch
from rest_framework import exceptions
from rest_framework.exceptions import AuthenticationFailed

from django.contrib.auth.models import Group
from django_entra_auth.config import ProviderConfig, Settings
from django_entra_auth.rest_framework import AdfsAccessTokenAuthentication
from .utils import (
    build_access_token_adfs,
    build_access_token_azure,
    build_access_token_azure_guest,
    build_access_token_azure_guest_no_upn,
    build_access_token_azure_not_guest,
    build_access_token_azure_guest_with_idp,
    build_access_token_azure_groups_in_claim_source,
    mock_adfs,
)


class RestFrameworkIntegrationTests(TestCase):
    def setUp(self):
        self.drf_auth_class = AdfsAccessTokenAuthentication()

        adfs_response = build_access_token_adfs(RequestFactory().get("/"))[2]
        self.access_token_adfs = json.loads(adfs_response)["access_token"]

        azure_response = build_access_token_azure(RequestFactory().get("/"))[2]
        self.access_token_azure = json.loads(azure_response)["access_token"]

        azure_response_guest = build_access_token_azure_guest(
            RequestFactory().get("/")
        )[2]
        self.access_token_azure_guest = json.loads(azure_response_guest)["access_token"]

        azure_response_no_guest = build_access_token_azure_not_guest(
            RequestFactory().get("/")
        )[2]
        self.access_token_azure_no_guest = json.loads(azure_response_no_guest)[
            "access_token"
        ]

        azure_response_guest = build_access_token_azure_guest_no_upn(
            RequestFactory().get("/")
        )[2]
        self.access_token_azure_guest_no_upn = json.loads(azure_response_guest)[
            "access_token"
        ]

        azure_response_guest = build_access_token_azure_guest_with_idp(
            RequestFactory().get("/")
        )[2]
        self.access_token_azure_guest_with_idp = json.loads(azure_response_guest)[
            "access_token"
        ]

        azure_response = build_access_token_azure_groups_in_claim_source(
            RequestFactory().get("/")
        )[2]
        self.access_token_azure_groups_in_claim_source = json.loads(azure_response)[
            "access_token"
        ]

        Group.objects.create(name="group1")
        Group.objects.create(name="group2")
        Group.objects.create(name="group3")

    @mock_adfs("azure")
    def test_access_token_azure(self):
        access_token_header = "Bearer {}".format(self.access_token_azure)
        request = RequestFactory().get("/api", HTTP_AUTHORIZATION=access_token_header)

        from django_entra_auth.config import django_settings

        settings = deepcopy(django_settings)
        # SERVER is no longer needed as it's now fixed to login.microsoftonline.com
        settings.ENTRA_AUTH["TENANT_ID"] = "dummy_tenant_id"
        with patch("django_entra_auth.config.django_settings", settings):
            with patch("django_entra_auth.config.settings", Settings()):
                with patch(
                    "django_entra_auth.backend.provider_config", ProviderConfig()
                ):
                    user, token = self.drf_auth_class.authenticate(request)
                    self.assertEqual(user.username, "testuser")

    @mock_adfs("azure")
    def test_access_token_azure_guest(self):
        access_token_header = "Bearer {}".format(self.access_token_azure_guest)
        request = RequestFactory().get("/api", HTTP_AUTHORIZATION=access_token_header)

        from django_entra_auth.config import django_settings

        settings = deepcopy(django_settings)
        # SERVER is no longer needed as it's now fixed to login.microsoftonline.com
        settings.ENTRA_AUTH["TENANT_ID"] = "dummy_tenant_id"
        settings.ENTRA_AUTH["BLOCK_GUEST_USERS"] = True
        with patch("django_entra_auth.config.django_settings", settings):
            with patch("django_entra_auth.backend.settings", Settings()):
                with patch("django_entra_auth.config.settings", Settings()):
                    with patch(
                        "django_entra_auth.backend.provider_config", ProviderConfig()
                    ):
                        with self.assertRaises(AuthenticationFailed):
                            user, token = self.drf_auth_class.authenticate(request)

    @mock_adfs("azure")
    def test_access_token_azure_no_guest(self):
        access_token_header = "Bearer {}".format(self.access_token_azure_no_guest)
        request = RequestFactory().get("/api", HTTP_AUTHORIZATION=access_token_header)

        from django_entra_auth.config import django_settings

        settings = deepcopy(django_settings)
        # SERVER is no longer needed as it's now fixed to login.microsoftonline.com
        settings.ENTRA_AUTH["TENANT_ID"] = "dummy_tenant_id"
        settings.ENTRA_AUTH["BLOCK_GUEST_USERS"] = True
        with patch("django_entra_auth.config.django_settings", settings):
            with patch("django_entra_auth.backend.settings", Settings()):
                with patch("django_entra_auth.config.settings", Settings()):
                    with patch(
                        "django_entra_auth.backend.provider_config", ProviderConfig()
                    ):
                        user, token = self.drf_auth_class.authenticate(request)
                        self.assertEqual(user.username, "testuser")

    @mock_adfs("azure")
    def test_access_token_azure_guest_but_no_upn(self):
        access_token_header = "Bearer {}".format(self.access_token_azure_guest_no_upn)
        request = RequestFactory().get("/api", HTTP_AUTHORIZATION=access_token_header)
        from django_entra_auth.config import django_settings

        settings = deepcopy(django_settings)
        # SERVER is no longer needed as it's now fixed to login.microsoftonline.com
        settings.ENTRA_AUTH["TENANT_ID"] = "dummy_tenant_id"
        settings.ENTRA_AUTH["GUEST_USERNAME_CLAIM"] = "email"
        settings.ENTRA_AUTH["BLOCK_GUEST_USERS"] = False
        with patch("django_entra_auth.config.django_settings", settings):
            with patch("django_entra_auth.backend.settings", Settings()):
                with patch("django_entra_auth.config.settings", Settings()):
                    with patch(
                        "django_entra_auth.backend.provider_config", ProviderConfig()
                    ):
                        user, token = self.drf_auth_class.authenticate(request)
                        self.assertEqual(user.username, "john.doe@example.com")

    @mock_adfs("azure")
    def test_access_token_azure_guest_with_idp(self):
        access_token_header = "Bearer {}".format(self.access_token_azure_guest_with_idp)
        request = RequestFactory().get("/api", HTTP_AUTHORIZATION=access_token_header)
        from django_entra_auth.config import django_settings

        settings = deepcopy(django_settings)
        # SERVER is no longer needed as it's now fixed to login.microsoftonline.com
        settings.ENTRA_AUTH["TENANT_ID"] = "dummy_tenant_id"
        settings.ENTRA_AUTH["GUEST_USERNAME_CLAIM"] = "email"
        settings.ENTRA_AUTH["BLOCK_GUEST_USERS"] = False
        with patch("django_entra_auth.config.django_settings", settings):
            with patch("django_entra_auth.backend.settings", Settings()):
                with patch("django_entra_auth.config.settings", Settings()):
                    with patch(
                        "django_entra_auth.backend.provider_config", ProviderConfig()
                    ):
                        user, token = self.drf_auth_class.authenticate(request)
                        self.assertEqual(user.username, "john.doe@example.com")

    @mock_adfs("azure")
    def test_access_token_azure_guest_but_no_upn_but_no_guest_username_claim(self):
        access_token_header = "Bearer {}".format(self.access_token_azure_guest_no_upn)
        request = RequestFactory().get("/api", HTTP_AUTHORIZATION=access_token_header)
        from django_entra_auth.config import django_settings

        settings = deepcopy(django_settings)
        # SERVER is no longer needed as it's now fixed to login.microsoftonline.com
        settings.ENTRA_AUTH["TENANT_ID"] = "dummy_tenant_id"
        settings.ENTRA_AUTH["GUEST_USERNAME_CLAIM"] = (
            None  # <--- Set to None, should not be validated as OK
        )
        settings.ENTRA_AUTH["BLOCK_GUEST_USERS"] = False
        with patch("django_entra_auth.config.django_settings", settings):
            with patch("django_entra_auth.backend.settings", Settings()):
                with patch("django_entra_auth.config.settings", Settings()):
                    with patch(
                        "django_entra_auth.backend.provider_config", ProviderConfig()
                    ):
                        with self.assertRaises(exceptions.AuthenticationFailed):
                            self.drf_auth_class.authenticate(request)

    @mock_adfs("azure", requires_obo=True)
    def test_process_group_claim_from_ms_graph(self):
        access_token_header = "Bearer {}".format(
            self.access_token_azure_groups_in_claim_source
        )
        request = RequestFactory().get("/api", HTTP_AUTHORIZATION=access_token_header)

        from django_entra_auth.config import django_settings

        settings = deepcopy(django_settings)
        # SERVER is no longer needed as it's now fixed to login.microsoftonline.com
        settings.ENTRA_AUTH["TENANT_ID"] = "dummy_tenant_id"
        with patch("django_entra_auth.config.django_settings", settings):
            with patch("django_entra_auth.backend.settings", Settings()):
                with patch("django_entra_auth.config.settings", Settings()):
                    with patch(
                        "django_entra_auth.backend.provider_config", ProviderConfig()
                    ):
                        user, _ = self.drf_auth_class.authenticate(request)
                        self.assertEqual(user.username, "testuser")
                        self.assertEqual(user.groups.all()[0].name, "group1")
                        self.assertEqual(user.groups.all()[1].name, "group2")

    @mock_adfs("azure", requires_obo=True, mfa_error=True)
    def test_get_obo_access_token_mfa_error(self):
        access_token_header = "Bearer {}".format(
            self.access_token_azure_groups_in_claim_source
        )
        request = RequestFactory().get("/api", HTTP_AUTHORIZATION=access_token_header)

        from django_entra_auth.config import django_settings

        settings = deepcopy(django_settings)
        # SERVER is no longer needed as it's now fixed to login.microsoftonline.com
        settings.ENTRA_AUTH["TENANT_ID"] = "dummy_tenant_id"
        with patch("django_entra_auth.config.django_settings", settings):
            with patch("django_entra_auth.backend.settings", Settings()):
                with patch("django_entra_auth.config.settings", Settings()):
                    with patch(
                        "django_entra_auth.backend.provider_config", ProviderConfig()
                    ):
                        with self.assertRaises(AuthenticationFailed):
                            self.drf_auth_class.authenticate(request)

    @mock_adfs("azure", requires_obo=True, version="v2.0")
    def test_get_obo_access_token_version_2(self):
        access_token_header = "Bearer {}".format(
            self.access_token_azure_groups_in_claim_source
        )
        request = RequestFactory().get("/api", HTTP_AUTHORIZATION=access_token_header)

        from django_entra_auth.config import django_settings

        settings = deepcopy(django_settings)
        # SERVER is no longer needed as it's now fixed to login.microsoftonline.com
        settings.ENTRA_AUTH["TENANT_ID"] = "dummy_tenant_id"
        settings.ENTRA_AUTH["VERSION"] = "v2.0"
        with patch("django_entra_auth.config.django_settings", settings):
            with patch("django_entra_auth.backend.settings", Settings()):
                with patch("django_entra_auth.config.settings", Settings()):
                    with patch(
                        "django_entra_auth.backend.provider_config", ProviderConfig()
                    ):
                        user, _ = self.drf_auth_class.authenticate(request)
                        self.assertEqual(user.username, "testuser")
                        self.assertEqual(user.groups.all()[0].name, "group1")
                        self.assertEqual(user.groups.all()[1].name, "group2")

    @mock_adfs("azure", requires_obo=True, missing_graph_group_perm=True)
    def test_missing_ms_graph_group_permission(self):
        access_token_header = "Bearer {}".format(
            self.access_token_azure_groups_in_claim_source
        )
        request = RequestFactory().get("/api", HTTP_AUTHORIZATION=access_token_header)

        from django_entra_auth.config import django_settings

        settings = deepcopy(django_settings)
        # SERVER is no longer needed as it's now fixed to login.microsoftonline.com
        settings.ENTRA_AUTH["TENANT_ID"] = "dummy_tenant_id"
        with patch("django_entra_auth.config.django_settings", settings):
            with patch("django_entra_auth.backend.settings", Settings()):
                with patch("django_entra_auth.config.settings", Settings()):
                    with patch(
                        "django_entra_auth.backend.provider_config", ProviderConfig()
                    ):
                        with self.assertRaises(AuthenticationFailed):
                            self.drf_auth_class.authenticate(request)
